# -*- coding: utf-8 -*-
#
# This file is part of SIDEKIT.
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#
# SIDEKIT is free software: you can redistribute it and/or modify
# it under the terms of the GNU LLesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# SIDEKIT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.

"""
Copyright 2014-2021 Anthony Larcher & Martin Lebourdais & Theo Mariotte

"""

import h5py
import numpy
import pandas
import pickle
import shelve
import time
import torch
import torchaudio
import tqdm

from ..diar import Diar
from sidekit.nnet.augmentation import data_augmentation
from datetime import datetime


__license__ = "LGPL"
__author__ = "Theo Mariotte & Anthony Larcher"
__copyright__ = "Copyright 2015-2021 Anthony Larcher"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reStructuredText'


class SeqSetRandomSampler(torch.utils.data.Sampler):
    """
    Random data sampler used to build batches in PyTprch DataLoader
    This sampler randomly selects training segments in the database that will be used for training.

    """
    def __init__(self, batch_size, batch_num, list_file, seg_set,rng=None, seed=1234):
        """

        :param batch_size: number of segments per batch
        :param batch_num: number of batches per epoch of training
        :param list_file: list of the files to be used for the training
        :param seg_set: segment sets previously generated using data preparation
        :param rng: random number generator (default : None)
        :param seed: seed of the random generator (default : 1234)
        """
        if rng is None:
            self.rng = numpy.random.default_rng(seed)
        else:
            self.rng = rng
        with open(list_file, "r") as ff:
            self.show_list = [l.strip() for l in ff.readlines()]

        idx_iter = list()
        segments = shelve.open(seg_set)
        for show in tqdm.tqdm(
                self.show_list, desc="Sampler initialization,first pass", unit="show"
        ):
            seg_tags = segments.get(show)

            for seg in seg_tags:
                idx_iter.append({'seg':seg,'overlap':None})

        self.index_iterator = numpy.array(idx_iter)
        self.length = batch_num * batch_size
        segments.close()

    def __iter__(self):
        """

        :return:
        """
        self.rng.shuffle(self.index_iterator)
        self.iter = self.index_iterator[: self.length]
        return iter(self.iter)

    def __len__(self) -> int:
        """

        :return:
        """
        return self.length


class SeqSetSampler(torch.utils.data.Sampler):
    """

    """
    def __init__(self,
                 list_file,
                 seg_set,
                 speech_ratio,
                 spk_turn_ratio,
                 overlap_ratio,
                 artificial_overlap_ratio,
                 batch_size=64,
                 batch_number=None,
                 task="overlap",
                 rng=None,
                 seed=1234):
        """
        Fast Sampler from pytorch that respect contraints:
            - Not more than one segment per show in a batch (as much as possible)
            - Ratio of overlap segments, speech segment, speaker turn segment

        :param show_list: list of the shows on the training set
        :param seg_set: path to the shelve file where segments and tags are stored
        :param speech_ratio: ratio of segments annotated as speech only (between 0 and 1)
        :param spk_turn_ratio: ratio of segments annotated as speech turn (between 0 and 1)
        :param overlap: ratio of segments annotated as overlaps (between 0 and 1)
        :param batch_size: size of the batch
        :param batch_number: number of batches to be generated
        :param task: Task the sampler is needed for
        :param rng: random number generator for Numpy
        :param seed: seed for random generator
        """
        # List of Uris
        with open(list_file, "r") as ff:
            self.show_list = [l.strip() for l in ff.readlines()]
        if rng is None:
            self.rng = numpy.random.default_rng(seed)
        else:
            self.rng = rng
        self.speech_ratio = speech_ratio
        self.spkT_ratio = spk_turn_ratio
        self.overlap_ratio = overlap_ratio
        self.artificial_overlap_ratio = artificial_overlap_ratio
        self.batch_size = batch_size

        batch_num_max = 10000
        self.batch_num = batch_num_max if (batch_number is None) else batch_number
        # dictionnary containing global indexes of seglents for each show
        self.seg_per_show = dict()

        # open shelve file containing segments start times for each show + segments tags
        # (i.e. speech/non-speech ; spk turn ; overlap)
        # Index of each segment w.r.t. the whole dataset are also saved here
        segments = shelve.open(seg_set)
        len_ov = []
        len_spkT = []
        len_speech = []
        len_nospeech = []
        self.shows = {
            "overlap": {"list": [], "probs": []},
            "speaker_change": {"list": [], "probs": []},
            "speech": {"list": [], "probs": []},
            "nonspeech": {"list": [], "probs": []},
        }

        for show in tqdm.tqdm(
                self.show_list, desc="Sampler initialization,first pass", unit="show"
        ):

            # Classify each segment (in terms of global index)
            # for speaker_change and overlap you'll need an other list
            idx_speech = list()
            idx_nonspeech = list()
            idx_spkT = list()
            idx_overlap = list()
            lst_speech = []
            lst_nonspeech = []
            lst_spkT = []
            lst_overlap = []

            self.seg_tags = segments.get(show)
            #test
            # print(self.seg_tags)
            # return

            for ii in range(len(self.seg_tags)):
                seg = self.seg_tags[ii]

                # segment labelled as overlap
                if seg["overlap"]:
                    lst_overlap.append(seg)
                # segment labelled as speech turn
                elif seg["speaker_change"]:
                    lst_spkT.append(seg)
                # segment labelled as speech/non-speech
                elif seg["speech"]:
                    lst_speech.append(seg)
                else:
                    lst_nonspeech.append(seg)

            if lst_overlap:
                self.shows["overlap"]["list"].append(show)
            if lst_spkT:
                self.shows["speaker_change"]["list"].append(show)
            if lst_speech:
                self.shows["speech"]["list"].append(show)
            if lst_nonspeech:
                self.shows["nonspeech"]["list"].append(show)

            len_ov.append(len(lst_overlap))
            len_spkT.append(len(lst_spkT))
            len_speech.append(len(lst_speech))
            len_nospeech.append(len(lst_nonspeech))

            # all indexes are stored in a dictionary with each index classified
            self.seg_per_show[show] = {
                "overlap": numpy.array(lst_overlap),
                "speaker_change": numpy.array(lst_spkT),
                "speech": numpy.array(lst_speech),
                "nonspeech": numpy.array(lst_nonspeech),
            }

        # max length of the limiting segment without a possible outlier
        len_max_ov = max(int(numpy.mean(len_ov) + 2 * numpy.std(len_ov)),self.batch_num)

        # check list sizes w.r.t. the batch size
        if len(self.shows["overlap"]["list"]) < self.batch_size:
            self.shows["overlap"]["list"] = self.shows["overlap"]["list"] * numpy.ceil(
                self.batch_size / len(self.shows["overlap"]["list"])
            ).astype("int")

        if len(self.shows["speaker_change"]["list"]) < self.batch_size:
            self.shows["speaker_change"]["list"] = self.shows["speaker_change"]["list"] * numpy.ceil(
                self.batch_size / len(self.shows["speaker_change"]["list"])
            ).astype("int")

        if len(self.shows["speech"]["list"]) < self.batch_size:
            self.shows["speech"]["list"] = self.shows["speech"]["list"] * numpy.ceil(
                self.batch_size / len(self.shows["speech"]["list"])
            ).astype("int")

        if len(self.shows["nonspeech"]["list"]) < self.batch_size:
            self.shows["nonspeech"]["list"] = self.shows["nonspeech"]["list"] * numpy.ceil(
                self.batch_size / len(self.shows["nonspeech"]["list"])
            ).astype("int")

        # probability vectors
        self.shows["overlap"]["probs"] = numpy.ones(len(self.shows["overlap"]["list"]), dtype=int)  # will become a list
        self.shows["speaker_change"]["probs"] = []
        self.shows["speech"]["probs"] = []
        self.shows["nonspeech"]["probs"] = []

        self.batch_num = min(self.batch_num,batch_num_max)
        len_max_speech = max(int(numpy.mean(len_speech) + 2 * numpy.std(len_speech)),self.batch_num)
        len_max_nospeech = max(int(numpy.mean(len_nospeech) + 2 * numpy.std(len_nospeech)),self.batch_num)
        len_max_spkT = max(int(numpy.mean(len_spkT) + 2 * numpy.std(len_spkT)),self.batch_num)

        for show in tqdm.tqdm(
                self.show_list, desc="Sampler initialization, Padding", unit="show"
        ):
            if show in self.shows["overlap"]["list"]:
                # self.shows["overlap"]['probs'].append(len_max_unit)
                if 0 < len(self.seg_per_show[show]["overlap"]) < len_max_ov:

                    tmplist = numpy.repeat(
                        self.seg_per_show[show]["overlap"],
                        numpy.ceil(
                            len_max_ov / len(self.seg_per_show[show]["overlap"])
                        ).astype("int"),
                        axis=0
                    )
                else:
                    tmplist = self.seg_per_show[show]["overlap"]

                self.seg_per_show[show]["overlap"] = {
                    "data": self.rng.permutation(tmplist[:len_max_ov]),
                    "probs": numpy.ones(len_max_ov)
                }

            if show in self.shows["speaker_change"]["list"]:
                # self.shows["overlap"]['probs'].append(len_max_unit)
                if 0 < len(self.seg_per_show[show]["speaker_change"]) < len_max_spkT:

                    tmplist = numpy.repeat(
                        self.seg_per_show[show]["speaker_change"],
                        numpy.ceil(
                            len_max_spkT / len(self.seg_per_show[show]["speaker_change"])
                        ).astype("int"),
                        axis=0
                    )
                else:
                    tmplist = self.seg_per_show[show]["speaker_change"]

                self.seg_per_show[show]["speaker_change"] = {
                    "data": self.rng.permutation(tmplist[:len_max_spkT]),
                    "probs": numpy.ones(len_max_spkT),
                }

            if 0 < len(self.seg_per_show[show]["speech"]) < len_max_speech:
                try:
                    tmplist = numpy.repeat(
                        self.seg_per_show[show]["speech"],
                        numpy.ceil(
                            len_max_speech / len(self.seg_per_show[show]["speech"])
                        ).astype("int"),
                        axis=0
                    )
                except ZeroDivisionError as e:
                    print(show)
                    raise ZeroDivisionError(show)


            else:
                tmplist = self.seg_per_show[show]["speech"]

            self.seg_per_show[show]["speech"] = {
                "data": self.rng.permutation(tmplist[:len_max_speech]),
                "probs": numpy.ones(len_max_speech),
            }

            if 0 < len(self.seg_per_show[show]["nonspeech"]) < len_max_nospeech:

                tmplist = numpy.repeat(
                    self.seg_per_show[show]["nonspeech"],
                    numpy.ceil(
                        len_max_nospeech / len(self.seg_per_show[show]["nonspeech"])
                    ).astype("int"),
                    axis=0
                )

            else:
                tmplist = self.seg_per_show[show]["nonspeech"]

            self.seg_per_show[show]["nonspeech"] = {
                "data": self.rng.permutation(tmplist[:len_max_nospeech]),
                "probs": numpy.ones(len_max_nospeech),
            }

        for show in self.shows["speaker_change"]['list']:
            self.shows["speaker_change"]["probs"].append(
                len(self.seg_per_show[show]["speaker_change"]["data"])
            )

        for show in self.shows["speech"]['list']:
            self.shows["speech"]["probs"].append(
                len(self.seg_per_show[show]["speech"]["data"])
            )

        for show in self.shows["nonspeech"]['list']:
            self.shows["nonspeech"]["probs"].append(
                len(self.seg_per_show[show]["nonspeech"]["data"])
            )

        self.shows["overlap"]["probs"] *= len_max_ov
        self.shows["speaker_change"]["probs"] = numpy.array(self.shows["speaker_change"]["probs"])
        self.shows["speech"]["probs"] = numpy.array(self.shows["speech"]["probs"])
        self.shows["nonspeech"]["probs"] = numpy.array(self.shows["nonspeech"]["probs"])

        self.shows["speaker_change"]["probs"] = self.shows["speaker_change"]["probs"][
            self.shows["speaker_change"]["probs"] > 0
            ]
        self.shows["speech"]["probs"] = self.shows["speech"]["probs"][
            self.shows["speech"]["probs"] > 0
            ]
        self.shows["nonspeech"]["probs"] = self.shows["nonspeech"]["probs"][
            self.shows["nonspeech"]["probs"] > 0
            ]

        self.shows["overlap"]["list"] = numpy.array(self.shows["overlap"]["list"])
        self.shows["speaker_change"]["list"] = numpy.array(self.shows["speaker_change"]["list"])
        self.shows["speech"]["list"] = numpy.array(self.shows["speech"]["list"])
        self.shows["nonspeech"]["list"] = numpy.array(self.shows["nonspeech"]["list"])
        self.length = self.batch_size * self.batch_num
        self.backup_shows = pickle.dumps(self.shows)
        self.backup_seg_per_show = pickle.dumps(self.seg_per_show)
        segments.close()
        self.index_iterator = None

    def __iter__(self):
        """

        :return:
        """
        # count the number of segment of each type
        if self.index_iterator is None:
            N_ola = int(self.overlap_ratio * self.batch_size)
            N_spkT = int(self.spkT_ratio * self.batch_size)
            N_speech = int(self.speech_ratio * self.batch_size)
            N_artificial_ov = int(self.artificial_overlap_ratio * self.batch_size)
            # In case of overlap
            N_nospeech = max(0,self.batch_size - (N_ola + N_speech + N_artificial_ov + N_spkT))

            index_iterator = []

            # RAZ

            t0 = time.time()
            self.seg_per_show = pickle.loads(self.backup_seg_per_show)
            self.shows = pickle.loads(self.backup_shows)

            #self.index_iterator = numpy.zeros((self.length,), dtype=int)

            for batch_idx in range(self.batch_num-1):#Last one fail

                # Overlap segment selection

                probs = self.shows["overlap"]["probs"]
                N = len(self.shows["overlap"]["list"])

                ov_show_idx = self.rng.choice(
                    N, size=N_ola, replace=False, p=probs / numpy.sum(probs)
                ).astype("int")
                self.shows["overlap"]["probs"][ov_show_idx] -= 1

                ov_show = self.shows["overlap"]["list"][ov_show_idx]

                # Speaker turn segments selection
                probs = self.shows["speaker_change"]["probs"]
                N = len(self.shows["speaker_change"]["list"])

                spkT_show_idx = self.rng.choice(
                    N, size=N_spkT, replace=False, p=probs / numpy.sum(probs)
                ).astype("int")
                self.shows["speaker_change"]["probs"][spkT_show_idx] -= 1

                spkT_show = self.shows["speaker_change"]["list"][spkT_show_idx]

                # Speech segment selection
                probs = self.shows["speech"]["probs"]
                N = len(self.shows["speech"]["list"])
                speech_show_idx = self.rng.choice(
                    N, size=N_speech, replace=False, p=probs / numpy.sum(probs)
                ).astype("int")
                self.shows["speech"]["probs"][speech_show_idx] -= 1
                speech_show = self.shows["speech"]["list"][speech_show_idx]

                #Artificial overlap selection

                N = len(self.shows["speech"]["list"])
                art_ov_show_idx = self.rng.choice(
                    N, size=N_artificial_ov, replace=False,
                ).astype("int")

                art_ov_show = self.shows["speech"]["list"][art_ov_show_idx]

                # Non speech segment selection
                probs = self.shows["nonspeech"]["probs"]
                N = len(self.shows["nonspeech"]["list"])

                nospeech_show_idx = self.rng.choice(
                    N, size=N_nospeech, replace=False, p=probs / numpy.sum(probs)
                ).astype("int")
                self.shows["nonspeech"]["probs"][nospeech_show_idx] -= 1
                nospeech_show = self.shows["nonspeech"]["list"][nospeech_show_idx]

                batch_ = []
                ii = 0
                for show in ov_show:
                    probs = self.seg_per_show[show]["overlap"]["probs"]

                    seg_idx = int(
                        self.rng.choice(
                            len(self.seg_per_show[show]["overlap"]["data"]),
                            p=probs / numpy.sum(probs),
                        )
                    )  # only one value so can't use astype

                    self.seg_per_show[show]["overlap"]["probs"][seg_idx] -= 1

                    structure = {
                        'seg':self.seg_per_show[show]["overlap"]["data"][seg_idx],
                        'overlap':None
                    }
                    batch_.append(structure)

                for show in spkT_show:
                    probs = self.seg_per_show[show]["speaker_change"]["probs"]

                    seg_idx = int(
                        self.rng.choice(
                            len(self.seg_per_show[show]["speaker_change"]["data"]),
                            p=probs / numpy.sum(probs),
                        )
                    )  # only one value so can't use astype

                for show in speech_show:
                    probs = self.seg_per_show[show]["speech"]["probs"]
                    seg_idx = int(
                        self.rng.choice(
                            len(self.seg_per_show[show]["speech"]["data"]),
                            p=probs / numpy.sum(probs),
                        )
                    )
                    self.seg_per_show[show]["speech"]["probs"][seg_idx] -= 1
                    structure = {
                        'seg':self.seg_per_show[show]["speech"]["data"][seg_idx],
                        'overlap':None
                    }
                    batch_.append(structure)


                for show in art_ov_show:

                    seg_idx = int(
                        self.rng.choice(
                            len(self.seg_per_show[show]["speech"]["data"]),

                        )
                    )
                    ov_seg_idx = int(
                        self.rng.choice(
                            len(self.seg_per_show[show]["speech"]["data"]),

                        )
                    )
                    structure = {
                        'seg':self.seg_per_show[show]["speech"]["data"][seg_idx],
                        'overlap':self.seg_per_show[show]["speech"]["data"][ov_seg_idx]
                    }
                    batch_.append(structure)

                for show in nospeech_show:
                    probs = self.seg_per_show[show]["nonspeech"]["probs"]
                    seg_idx = int(
                        self.rng.choice(
                            len(self.seg_per_show[show]["nonspeech"]["data"]),
                            p=probs / numpy.sum(probs),
                        )
                    )
                    self.seg_per_show[show]["nonspeech"]["probs"][seg_idx] -= 1
                    structure = {
                        'seg':self.seg_per_show[show]["nonspeech"]["data"][seg_idx],
                        'overlap':None
                    }
                    batch_.append(structure)
                self.rng.shuffle(batch_)

                index_iterator = index_iterator + batch_
            # self.index_iterator = index_iterator
            self.index_iterator = None

        else:
            index_iterator = self.index_iterator

        return iter(index_iterator)

    def __len__(self) -> int:
        return self.length


class SeqSet(torch.utils.data.Dataset):
    """
    DataSet for sequence to sequence model training
    """

    def __init__(self, dataset_params, mode="train", logger=None, rng=None, seed=1234):
        """
        constructor

        :param dataset_params: parameters of the dataset extracted from YAML file (dictionnary)
        :param mode: train or evaluation mode (default="train")
        :param logger: logging object to save dataset parameters
        :param rng: random number generator for Numpy (default None)
        :param seed: random seed to control rng
        """
        torchaudio.set_audio_backend("soundfile")
        if rng is None:
            self.rng = numpy.random.default_rng(seed)
        else:
            self.rng = rng
        self.wav_dir = dataset_params["wav_dir"]
        self.label_set = dataset_params[mode]["label_set"]
        self.seg_set = dataset_params[mode]["seg_set"]
        self.audio_fr = dataset_params["audio_fr"]
        self.output_fr = dataset_params["output_fr"]
        self.task = dataset_params["task"]
        self.hdf5_dataset = self.label_set
        self.babble_noise = dataset_params["babble_noise"]
        self.len_ = dataset_params["sampler"]["batch_number"]*dataset_params["sampler"]["batch_size"]
        self.time_base_start = dict()
        self.logger = logger
        with shelve.open(self.seg_set,'r') as segments:
            for show in segments:
                crnt_set = segments.get(show)
                self.time_base_start[show] = crnt_set[0]["start"] #centisenconds
        # labels augmentation for speaker turn detection
        if not "labelling" in dataset_params.keys():
            self.collar_duration = 0.125
        else:
            self.collar_duration = dataset_params["labelling"]["collar_duration"]

        # lire les fichiers contenant les t_start par show pour construire une liste de segments
        self.duration = numpy.ceil(
            dataset_params[mode]["duration"] * self.audio_fr
        ).astype(
            "int"
        )  # in samples

        # start of each file with respect to UEMS (to guarantee sync between raw labels and segment frontiers)
        self.time_base_start = dict()
        with shelve.open(self.seg_set,'r') as segments:
            for show in segments:
                crnt_set = segments.get(show)
                self.time_base_start[show] = crnt_set[0]["start"] #centisenconds

        if mode == "train":
            self.transformation = dataset_params["train"]["transformation"]
        else:
            self.transformation = dataset_params["eval"]["transformation"]

        self.transform = dict()
        if (self.transformation["pipeline"] != "") and (
                self.transformation["pipeline"] is not None
        ):
            print('Transformation required')
            self.transform_number = 1
            transforms = self.transformation
            self.noise_df = None
            if "AddNoise" in transforms:
                self.transform["add_noise"] = self.transformation["AddNoise"]

                noise_df = pandas.read_csv(self.transform["add_noise"]["noise_db_csv"])
                noise_df = noise_df.loc[noise_df.duration > dataset_params[mode]["duration"]]
                self.noise_df = noise_df.set_index(noise_df.type)

            self.rir_df = None
            if "AddReverb" in transforms:
                self.transform["add_reverb"] = self.transformation["AddReverb"]

                tmp_rir_df = pandas.read_csv(self.transform["add_reverb"]["rir_db_csv"])
                tmp_rir_df = tmp_rir_df.loc[tmp_rir_df["type"] == "simulated_rirs"]
                # load the RIR database
                self.rir_df = tmp_rir_df.set_index(tmp_rir_df.type)

    def __getitem__(self, struct):
        """
        Return a given audio segment and its associated labels at index in the dataset.
        Here, raw audio is returned (e.g. sampled at @16000 Hz or @8000 Hz). Labels are
        sampled at @100Hz since it's the model output frame rate.
        """

        # récupère les frontières des segments
        segarr = struct['seg']
        seg = {
            "start":float(segarr[3]),
            "show":segarr[0],
            "stop":float(segarr[4]),
        }

        idx_start = numpy.round(seg["start"] / 100.0 * self.audio_fr).astype(int)

        # load audio waveform (dim = (channels,length))
        # normalization is applied here
        waveform, speech_fs = torchaudio.load(
            filepath=self.wav_dir + seg["show"] + ".wav",
            frame_offset=idx_start,
            num_frames=self.duration,
            channels_first=True,
            normalize=True
        )
        #waveform = normalize(waveform)
        # is the signal mono or not ?
        is_multichannel = waveform.shape[0] > 1

        # add low energy noise to avoid zero values
        waveform += 1e-6 * torch.randn(waveform.shape[0],waveform.shape[1])

        # data augmentation if needed

        if self.transform and not is_multichannel:
            waveform = data_augmentation(
                waveform,
                speech_fs,
                self.transform,
                self.transform_number,
                noise_df=self.noise_df,
                rir_df=self.rir_df,
                babble_noise = self.babble_noise
            )
            if waveform is None:
                print(seg["show"],seg["start"] / 100.0,seg["stop"] / 100.0)
        # in case of multichannel signal, each channel has to be transformed
        elif self.transform and is_multichannel:
            for ii in range(waveform.shape[0]):
                waveform[ii,:] = data_augmentation(
                    waveform[ii,:],
                    speech_fs,
                    self.transform,
                    self.transform_number,
                    noise_df=self.noise_df,
                    rir_df=self.rir_df,
                    babble_noise = self.babble_noise
                )

        # get labels associated with the current audio training segment
        start = numpy.round(seg["start"] / 100.0 * self.output_fr).astype(int) - self.time_base_start[seg["show"]]
        stop = numpy.round(seg["stop"] / 100.0 * self.output_fr).astype(int) - self.time_base_start[seg["show"]]
        with h5py.File(self.label_set, "r") as data:
            crnt_label = data[seg["show"]]["total"][:, start:stop]
        expected_frames_num = int(self.duration/self.audio_fr * self.output_fr)
        if crnt_label.shape[1]<expected_frames_num:
            crnt_label = numpy.pad(crnt_label,[(0,0),(0,expected_frames_num - crnt_label.shape[1])])
        if crnt_label.shape[1] > expected_frames_num:
            crnt_label = crnt_label[:,:expected_frames_num]

        if self.task == "vad":
            output_label = (crnt_label > 0).astype(numpy.long)

        # may probably be optimized...
        elif self.task == "spk_turn":
            label = numpy.zeros_like(crnt_label)
            label[:,:-1] = (numpy.abs(crnt_label[:,:-1] - crnt_label[:,1:]) > 0).astype(
                numpy.long
            )
            # Apply convolution to replace diracs by a chosen shape (gate or triangle)
            filter_sample = int(self.collar_duration * self.output_fr * 2 + 1)

            conv_filt = numpy.ones((filter_sample,))
            output_label = numpy.convolve(conv_filt, label.squeeze(), mode='same')
            output_label = (numpy.expand_dims(output_label,axis=0)>=1).astype(numpy.long)

        elif self.task == "overlap":

            if struct['overlap'] is not None:

                # Loading artificial overlap
                ov_segarr = struct['overlap']
                ov_seg = {
                    "start":float(ov_segarr[3]),
                    "show":ov_segarr[0],
                    "stop":float(ov_segarr[4]),
                }
                idx_start_ov = numpy.round(ov_seg["start"] / 100.0 * self.audio_fr).astype(int)
                frame_count_ov = self.duration

                waveform_ov, _ = torchaudio.load(
                    filepath=self.wav_dir + ov_seg["show"] + ".wav",
                    frame_offset=idx_start_ov,
                    num_frames=frame_count_ov,
                    channels_first=True,
                )

                start_ov = numpy.round(ov_seg["start"] / 100.0 * self.output_fr).astype(int)
                stop_ov = numpy.round(ov_seg["stop"] / 100.0 * self.output_fr).astype(int)
                with h5py.File(self.label_set, "r") as data:
                    label_ov = data[seg["show"]]["total"][:, start_ov:stop_ov]
                if label_ov.shape[1]<expected_frames_num:
                    label_ov = numpy.pad(label_ov,[(0,0),(0,expected_frames_num - label_ov.shape[1])])
                if label_ov.shape[1] > expected_frames_num:
                    label_ov = label_ov[:,:expected_frames_num]
                speech_power = waveform.norm(p=2)
                noise_power = waveform_ov.norm(p=2)
                snr_db = 10 * self.rng.random()+1
                snr = 10 ** (snr_db / 20)
                scale = snr * noise_power / speech_power

                # Sum with artificial segment
                crnt_label += label_ov
                waveform = (scale * waveform + waveform_ov) / 2

            else:
                pass
            # Binarize the label
            output_label = (crnt_label > 1).astype(numpy.long)
        else:
            raise NotImplementedError()

        if torch.isnan(waveform).any():
            print("Waveform NAN !!!")

        return waveform, torch.from_numpy(output_label).T

    def __len__(self):
        return self.len_


class SeqSetSingle(torch.utils.data.Dataset):
    """
    DataSet for evaluating a modelwhich returns a sliding window over one file
    """
    def __init__(self,
                 show,
                 wav_fn,
                 label_set=None,
                 uem=None,
                 audio_framerate=16000,
                 output_framerate=100,
                 duration=2.0,
                 task="overlap",
                 shift=0.5):
        """
        Ctor

        :param show: name of the show on which to run prediction
        :param wave_fn: name of the wavefile to be loaded
        :param seg_set: segments of the test set
        :param label set: labels of the test set
        :param audio_framerate: samplerate of the audio signal
        :param output_framerate: output sample rate
        :param duration: duration of the test segment
        :param tranformation_pipeline: pipeline of transfomations to be applied
        """
        super(SeqSetSingle, self).__init__()
        self.show = show
        self.audio_fr = audio_framerate
        self.output_fr = output_framerate

        self.segment_list = Diar()
        self.task=task
        self.wav_name = wav_fn
        self.duration = numpy.ceil(duration * self.audio_fr).astype("int")  # in samples
        if label_set is None:
            self.give_label = False
        else:
            self.give_label = True
            self.hdf5_dataset = label_set
        # TODO HANDLE UEM
        waveform, sr = torchaudio.load(
            filepath=self.wav_name,
        )
        self.length = waveform.shape[1]

        # start of each file with respect to UEMS (to guarantee sync between raw labels and segment frontiers)
        self.time_base_start = 0
        if uem is not None:
            self.time_base_start = Diar.read_uem(uem)[0]['start']

        self.shift = numpy.ceil(shift * self.audio_fr ).astype("int")
        start = 0
        # Sliding window
        for val in numpy.arange(start=start,stop = self.length, step = self.shift):
            seg_start = (val/self.audio_fr)
            seg_stop = (seg_start + duration)#second
            self.segment_list.append(
                show=show,  # WARNING : vérifier les unités!!!
                start=seg_start,
                stop=seg_stop,
                cluster="",
            )

    def __getitem__(self, index):
        """
        returns a raw waveform segment at given index with its associated labels
        """

        # récupère les frontières des segments
        seg = self.segment_list[index]
        idx_start = numpy.ceil(seg["start"] * self.audio_fr).astype("int")

        # charge le segment wav de N canaux (torchaudio.load)

        waveform, _ = torchaudio.load(
            filepath=self.wav_name,
            frame_offset=idx_start,
            num_frames=self.duration,
            channels_first=True
        )
        if waveform.shape[1]<self.duration:
            waveform = torch.nn.functional.pad(waveform,(0,self.duration - waveform.shape[1]),"constant",0)

        # add low energy noise to avoid zero values
        waveform += (1e-6 * torch.randn(waveform.shape[0],waveform.shape[1])).long()

        # récupère les labels associés au segment
        if self.give_label:
            start = numpy.round(seg["start"] * self.output_fr).astype(int) - self.time_base_start
            stop = numpy.round(seg["stop"] * self.output_fr).astype(int) - self.time_base_start

            with h5py.File(self.hdf5_dataset, "r") as data:
                crnt_label = data[seg["show"]]["total"][:, start:stop]

            expected_frames_num = int(self.duration/self.audio_fr * self.output_fr)
            if crnt_label.shape[1]<expected_frames_num:
                crnt_label = numpy.pad(crnt_label,[(0,0),(0,expected_frames_num - crnt_label.shape[1])])
            if crnt_label.shape[1] > expected_frames_num:
                crnt_label = crnt_label[:,:expected_frames_num]

            # crnt_label = numpy.ones((stop-start,))
            if self.task == "vad":
                output_label = (crnt_label > 0).astype(numpy.long)

            elif self.task == "spk_turn":
                output_label = (numpy.abs(crnt_label[:-1] - crnt_label[1:]) > 0).astype(
                    numpy.long
                )

            elif self.task == "overlap":
                output_label = (crnt_label > 1).astype(numpy.long)

            else:
                raise NotImplementedError()

            return index,waveform, torch.from_numpy(output_label).T
        else:
            return index, waveform

    def __len__(self):
        """

        :return:
        """
        return len(self.segment_list)



# -*- coding: utf-8 -*-
#
# This file is part of s4d.
#
# s4d is a python package for speaker diarization.
# Home page: http://www-lium.univ-lemans.fr/s4d/
#
# s4d is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# s4d is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with s4d.  If not, see <http://www.gnu.org/licenses/>.


"""
Copyright 2014-2021 Sylvain Meignier and Anthony Larcher
"""

from .clustering.hac_bic import HAC_BIC
from .clustering.hac_clr import HAC_CLR
from  .clustering.hac_iv import hac_iv, hac_iv_it

from .clustering.hac_utils import argmin
from .clustering.hac_utils import argmax
from .clustering.hac_utils import roll
from .clustering.hac_utils import stat_server_remove
from .clustering.hac_utils import scores_remove
from .clustering.hac_utils import scores2distance
from .clustering.hac_utils import stat_server_merge
from .clustering.hac_utils import bic_square_root

from .clustering.cc_iv import ConnectedComponent


from .model_iv import ModelIV

from .diar import Diar
from .segmentation import sanity_check
from .segmentation import bic_linear
from .segmentation import segmentation

from .viterbi import Viterbi

from .scoring import DER

from .nnet import SeqToSeq
from .nnet import SeqSetSingle
from .nnet import SeqSetRandomSampler
from .nnet import SeqSetSampler
from .nnet import SeqSet



__author__ = "Sylvain Meignier and Anthony Larcher"
__copyright__ = "Copyright 2014-2021 Sylvain Meignier and Anthony Larcher"
__license__ = "LGPL"
__maintainer__ = "Sylvain Meignier"
__email__ = "sylvain.meignierr@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reStructuredText'
__version__ = "0.2.3"



# -*- coding: utf-8 -*-
#
# This file is part of s4d.
#
# s4d is a python package for speaker diarization.
# Home page: http://www-lium.univ-lemans.fr/s4d/
#
# s4d is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# s4d is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with s4d.  If not, see <http://www.gnu.org/licenses/>.

"""
Copyright 2014-2021 Anthony Larcher

"""

import numpy
import shutil
import torch

from datetime import datetime
from .sequence_sets import SeqSet
from .sequence_sets import SeqSetSampler
from .sequence_sets import SeqSetRandomSampler
from .sequence_models import SeqToSeq

from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter

from sklearn.metrics import confusion_matrix

__license__ = "LGPL"
__author__ = "Anthony Larcher & Martin Lebourdais & Theo Mariotte"
__copyright__ = "Copyright 2015-2021 Anthony Larcher"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reStructuredText'


def save_checkpoint(state, is_best, filename='checkpoint.pth.tar', best_filename='model_best.pth.tar'):
    """
    :param state:
    :param is_best:
    :param filename:
    :param best_filename:
    :return:
    """
    torch.save(state, filename)
    if is_best:
        shutil.copyfile(filename, best_filename)


def prepare_loaders(dataset_yaml, logger=None, rng_=None, seed=1234):
    """
    Prepare Pytoch DataLoader for training and evaluation of seq2seq model

    :param dataset_yaml:
    :param logger:
    :param rng_:
    :param seed:
    :return:
    """
    if rng_ is None:
        rng = numpy.random.default_rng(seed)
    else:
        rng = rng_
    dataset = dataset_yaml

    # training loader with custom data sampler
    task = dataset["task"]
    sampler_typ = dataset["sampler_type"]
    file_list = dataset["train"]["file_list"]
    eval_file_list = dataset["eval"]["file_list"]
    seg_set = dataset["train"]["seg_set"]
    eval_seg_set = dataset["eval"]["seg_set"]
    batch_size = dataset["sampler"]["batch_size"]
    batch_num = dataset["sampler"]["batch_number"]
    ola_ratio = dataset["sampler"]["overlap_ratio"]
    art_ola_ratio = dataset["sampler"]["artificial_overlap_ratio"]
    spkT_ratio = dataset["sampler"]["speaker_turn_ratio"]
    speech_ratio = dataset["sampler"]["speech_ratio"]
    num_workers = dataset["num_workers"]

    if sampler_typ == "balanced":
        print(f"Overlap proportion = {ola_ratio}, speech proportion = {speech_ratio}, artoverlap proportion = {art_ola_ratio}")
        sampler = SeqSetSampler(list_file=file_list,
                                seg_set=seg_set,
                                speech_ratio=speech_ratio,
                                spk_turn_ratio=spkT_ratio,
                                overlap_ratio=ola_ratio,
                                artificial_overlap_ratio = art_ola_ratio,
                                batch_size=batch_size,
                                batch_number=batch_num,
                                task=task,
                                rng = rng)
    elif sampler_typ == "random":
        sampler = SeqSetRandomSampler(batch_size=batch_size,
                                      batch_num=batch_num,
                                      list_file=file_list,
                                      seg_set=seg_set,
                                      rng = rng)

    else:
        raise NotImplementedError()

    training_set = SeqSet(dataset_params=dataset, mode="train", logger=logger, rng = rng)

    training_loader = DataLoader(training_set,
                                 sampler = sampler,
                                 batch_size=batch_size,
                                 num_workers=num_workers)

    evaluation_set = SeqSet(dataset_params=dataset,mode="eval", logger=logger, rng=rng)
    eval_sampler = SeqSetRandomSampler(batch_size=batch_size,
                                       batch_num=batch_num,
                                       list_file=eval_file_list,
                                       seg_set=eval_seg_set,
                                       rng = rng)

    eval_loader = DataLoader(evaluation_set,
                             sampler=eval_sampler,
                             batch_size=batch_size,
                             num_workers=num_workers)

    return training_loader, eval_loader


def seqtoseq_train(training_loader,
                   validation_loader,
                   model,
                   task,
                   training_yaml,
                   logger,
                   rng_=None,
                   seed=1234):
    """
    Training loop for seqtoseq model used for VAD, overlap and speaker
    turn detection tasks

    :param training_loader: Pytorch DataLoader for training set
    :param validation_loader: Pytorch DataLoader for validation set
    :param model: model to be trained
    :param task: training task (e.g. 'vad')
    :param training_yaml: YAML file containing training parameters
    :param logger: TrainLogger object to follow training and track process
    """
    if rng_ is None:
        rng = numpy.random.default_rng(seed)
    else:
        rng = rng_
    # read learning parameters
    train_params = training_yaml

    if "model_name" in train_params.keys():
        model_name=train_params["model_name"]
    else:
        model_name=None

    epochs = train_params["number_of_epochs"]
    opt = train_params["optimizer"]
    lr = train_params["learning_rate"]
    momentum = train_params["momentum"]
    patience = train_params["patience"]
    seed = train_params["seed"]
    xent_weights = train_params["xent_weights"]
    now = datetime.now()
    time = now.strftime("%m-%d_%H-%M")
    checkpoint_model_name = train_params["checkpoint_name"]
    best_model_name = train_params["best_name"]
    multi_gpu = train_params["multi_gpu"]
    num_thread = train_params["num_thread"]
    log_interval = train_params["log_interval"]

    if "scheduler" in train_params:
        scheduler_name = train_params["scheduler"].get("type",'None')
    else:
        scheduler_name = 'None'
    tensorboard_logs = train_params.get("tensorboard_logs",None)

    if tensorboard_logs is not None:
        tensorboard_logs += time+'/'
    # current device
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    # in case of multi_gpu training
    if torch.cuda.device_count() > 1 and multi_gpu:
        print("Let's use", torch.cuda.device_count(), "GPUs!")
        model = torch.nn.DataParallel(model)
    else:
        print("Train on a single GPU")

    # push model to GPU
    model.to(device)

    """
    Set the training options
    """
    if opt == 'sgd':
        _optimizer = torch.optim.SGD
        _options = {'lr': lr, 'momentum': momentum}
    elif opt == 'adam':
        _optimizer = torch.optim.Adam
        _options = {'lr': lr}
    elif opt == 'rmsprop':
        _optimizer = torch.optim.RMSprop
        _options = {'lr': lr}

    params = [
        {
            'params': [
                param for name, param in model.named_parameters() if 'bn' not in name
            ]
        },
        {
            'params': [
                param for name, param in model.named_parameters() if 'bn' in name
            ],
            'weight_decay': 0
        },
    ]

    optimizer = _optimizer([{'params': model.parameters()},], **_options)
    if scheduler_name == "Plateau":
        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, 'min', verbose=True,patience = train_params["scheduler"]["patience"])
    if scheduler_name == "Cycle":
        scheduler = torch.optim.lr_scheduler.CyclicLR(optimizer,mode='triangular2',base_lr=train_params["scheduler"]["base_lr"],max_lr=train_params["scheduler"]["max_lr"],step_size_up=train_params["scheduler"]["step"])
    if scheduler_name == "None":
        scheduler = None
    best_fmeas = 0.0
    best_fmeas_epoch = 1
    curr_patience = patience

    # for tensorboard training monitoring
    writer = SummaryWriter(log_dir=tensorboard_logs)
    for epoch in range(1, epochs + 1):

        # Process one epoch and return the current model
        model, optimizer, scheduler = train_epoch(model,
                                                  training_loader,
                                                  optimizer,
                                                  scheduler=scheduler,
                                                  log_interval=log_interval,
                                                  epoch=epoch,
                                                  device=device,
                                                  xent_weights=xent_weights,
                                                  writer=writer,
                                                  logger=logger)

        # validation
        val_loss, accuracy, fmeas = eval_epoch(validation_loader,
                                               model,
                                               device,
                                               epoch,
                                               xent_weights=xent_weights,
                                               writer=writer,
                                               logger=logger)

        # Decrease learning rate according to the scheduler policy
        # scheduler.step(val_loss)
        logger.message(stream=f"Learning rate is {optimizer.param_groups[0]['lr']}")

        # remember best accuracy and save checkpoint
        is_best = fmeas > best_fmeas
        best_fmeas = max(fmeas,best_fmeas)

        if type(model) is SeqToSeq:
            save_checkpoint({
                'epoch': epoch,
                'model_state_dict': model.state_dict(),
                'optimizer_state_dict': optimizer.state_dict(),
                'accuracy': best_fmeas,
                'scheduler': scheduler
            }, is_best, filename=checkpoint_model_name+".pt", best_filename=best_model_name + '.pt')
        else:
            save_checkpoint({
                'epoch': epoch,
                'model_state_dict': model.module.state_dict(),
                'optimizer_state_dict': optimizer.state_dict(),
                'accuracy': best_fmeas,
                'scheduler': scheduler
            }, is_best, filename=checkpoint_model_name +".pt", best_filename=best_model_name + '.pt')

        if is_best:
            best_fmeas_epoch = epoch
            curr_patience = patience
        else:
            curr_patience -= 1

    logger.message(stream=f"Best accuracy {best_fmeas * 100.} obtained at epoch {best_fmeas_epoch}")


def train_epoch(model,
                training_loader,
                optimizer,
                device,
                epoch,
                log_interval,
                scheduler=None,
                xent_weights=[0.5,0.5],
                writer=None,
                logger=None):
    """
    Model training along one epoch

    :param model: model to be trained
    :param training_loader: train set loader
    :param optimizer: current optimizer to update system weights
    :param device: current device (GPUs,...)
    :param epoch: index of the current epoch (for logging)
    :param log_interval: number of batch to process between two logging outputs
    :param xent_weights: weights of the cross entropy loss
    :param writer: logging on tensorboard
    :param logger: logging in a log file
    """
    model.to(device)
    model.train()
    # criterion = torch.nn.CrossEntropyLoss(reduction='mean', weight=torch.FloatTensor(xent_weights).to(device))
    criterion = torch.nn.CrossEntropyLoss(reduction='mean')

    recall = 0.0
    precision = 0.0
    accuracy = 0.0
    mean_speech_ratio = 0.0
    macrotp = 0.0
    macrofp = 0.0
    macrotn = 0.0
    macrofn = 0.0
    loss_acc = 0.0
    count = 0
    sm = torch.nn.Softmax(dim=1)
    for batch_idx, (data, target) in enumerate(training_loader):

        # target = target.squeeze()
        output = model(data.to(device))

        # Output shape = (Batch,time,nclass)
        # Target shape = (Batch,time)
        reshaped_output = torch.flatten(output,start_dim=0,end_dim=1)
        reshaped_target = torch.flatten(target)

        loss = criterion(reshaped_output,reshaped_target.to(device)) # Want (batch , class, time) and (batch,time) or (seq,class) (seq)
        loss_acc += loss
        count+=1
        if scheduler is not None:
            if isinstance(scheduler,torch.optim.lr_scheduler.CyclicLR):
                scheduler.step()
            else:
                scheduler.step(loss)
        loss.backward()
        optimizer.step()
        optimizer.zero_grad()
        output = sm(reshaped_output)
        tn, fp, fn, tp = confusion_matrix(reshaped_target.cpu(),torch.argmax(output,1).cpu()).ravel()
        macrotp += tp
        macrotn += tn
        macrofp += fp
        macrofn += fn

        if writer:
            writer.add_scalar('Loss/Train',loss.item(),epoch*len(training_loader)+batch_idx)
            writer.add_scalar('lr',optimizer.param_groups[0]['lr'],epoch*len(training_loader)+batch_idx)

        if batch_idx % log_interval == 0:
            precision, recall, f_measure, accuracy = matrix2metrics(macrotp, macrotn, macrofp, macrofn)
            averaged_loss = loss_acc/count
            loss_acc = 0.0
            count = 0
            macrotp = 0.0
            macrofp = 0.0
            macrotn = 0.0
            macrofn = 0.0
            batch_size = target.shape[0]
            if precision != 0 or recall != 0:
                logger.addStep(
                            mode="train",
                            epoch=epoch,
                            batch_idx=batch_idx,
                            batch_num=len(training_loader),
                            loss=averaged_loss,
                            metrics={"acc": accuracy,
                                     "precision": precision,
                                     "recall": recall,
                                     "fmeas": f_measure})
            else:
                f_measure = 0
                logger.addStep(
                            mode="train",
                            epoch=epoch,
                            batch_idx=batch_idx,
                            batch_num=len(training_loader),
                            loss=averaged_loss,
                            metrics={"acc": accuracy,
                                     "precision": precision,
                                     "recall": recall})

            if writer:
                writer.add_scalar('Acc/Train', accuracy, (epoch-1)*len(training_loader)+batch_idx)
                writer.add_scalar('F-meas/Train', f_measure, (epoch-1)*len(training_loader)+batch_idx)

    return model, optimizer, scheduler


def eval_epoch(validation_loader,
               model,
               device,
               epoch,
               xent_weights=[0.5, 0.5],
               writer=None,
               logger=None):
    """

    :param validation_loader:
    :param model:
    :param device:
    :param epoch:
    :param xent_weights:
    :param writer:
    :param logger:
    :return:
    """
    model.to(device)
    model.eval()
    criterion = torch.nn.CrossEntropyLoss(reduction='mean')

    recall = 0.0
    precision = 0.0
    accuracy = 0.0
    loss = 0.0
    macrotp = 0.0
    macrotn = 0.0
    macrofp = 0.0
    macrofn = 0.0
    sm = torch.nn.Softmax(dim=1)
    length = len(validation_loader)
    with torch.no_grad():

        for batch_idx, (data,target) in enumerate(validation_loader):
            output = model(data.to(device))

            reshaped_output = torch.flatten(output,start_dim=0,end_dim=1)
            reshaped_target = torch.flatten(target)

            loss += criterion(reshaped_output,reshaped_target.to(device)) # Want (batch , class, time) and (batch,time)

            output = sm(reshaped_output)

            try:
                tn, fp, fn, tp = confusion_matrix(reshaped_target.cpu(),torch.argmax(output,1).cpu()).ravel()
            except:
                tn = 0
                fp = 0
                fn = 0
                tp = 0
            macrotp += tp
            macrotn += tn
            macrofp += fp
            macrofn += fn

        # Metric calculation from confusion matrix
        precision, recall, f_measure, accuracy = matrix2metrics(macrotp, macrotn, macrofp, macrofn)
        if precision != 0 or recall != 0:
            # logger
            logger.addStep(
                        mode="eval",
                        epoch=epoch,
                        loss=loss/length,
                        metrics={"acc" : accuracy,
                                 "precision" : precision,
                                 "recall" : recall,
                                 "fmeas" : f_measure})
        else:
            f_measure=0
            logger.addStep(
                        mode="eval",
                        epoch=epoch,
                        loss=loss/length,
                        metrics={"acc" : accuracy,
                                 "precision" : precision,
                                 "recall" : recall})

        # tensor board
        if writer:
            writer.add_scalar('Loss/Eval', loss/length, epoch)
            writer.add_scalar('Acc/Eval', accuracy, epoch)
            if f_measure:
                writer.add_scalar('F-meas/Eval', f_measure, epoch)

    return loss/length, accuracy, f_measure


def matrix2metrics(tp, tn, fp, fn):
    """

    :param tp:
    :param tn:
    :param fp:
    :param fn:
    :return:
    """
    eps = 1e-7
    prec = tp/(tp+fp+eps)
    rec = tp/(tp+fn+eps)
    acc = (tp+tn)/(tp+tn+fp+fn+eps)
    fmes = (2*tp)/(2*tp+fp+fn+eps)
    return prec, rec, fmes, acc




# -*- coding: utf-8 -*-
#
# This file is part of s4d.
#
# s4d is a python package for speaker diarization.
# Home page: http://www-lium.univ-lemans.fr/s4d/
#
# s4d is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# s4d is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with s4d.  If not, see <http://www.gnu.org/licenses/>.

"""
Copyright 2014-2021 Anthony Larcher & Martin Lebourdais & Theo Mariotte

"""

import numpy
import torch
import torchaudio

from .sequence_sets import SeqSetSingle
from .utils import multi_label_combination
from .utils import filter_vad
from .utils import filter_speaker_turn
from .utils import filter_overlap
from collections import OrderedDict
from torch.utils.data import DataLoader

__license__ = "LGPL"
__author__ = "Theo Mariotte & Anthony Larcher"
__copyright__ = "Copyright 2015-2021 Anthony Larcher"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reStructuredText'


class Deltas(torch.nn.Module):
    """
    Class to add deltas and delta deltas
    """
    def __init__(self,
                 e=False,
                 d=False,
                 de=True,
                 dd=False,
                 dde=True):
        """
        :param e : energy
        :param d : first delta
        :param de : delta energy
        :param dd : delta delta
        :param dde : delta delta energy
        """
        super(Deltas,self).__init__()
        self.e = e
        self.d = d
        self.dd = dd
        #if self.d or self.dd:
        self.delta = torchaudio.transforms.ComputeDeltas()

    def forward(self,coefs):
        coefs = coefs.squeeze(dim=1)
        out=coefs
        #Caculation
        if self.d:
            delt = self.delta(coefs)
        if self.dd:
            delt2 = self.delta(delt)

        #concat
        if not self.e:
            coefs = coefs[:,1:,:]
            out=coefs
        if self.d:
            out = torch.cat((out,delt),dim=1)
        if self.dd:
            out = torch.cat((out,delt2),dim=1)
        return out


class BLSTM(torch.nn.Module):
    """
    Bi LSTM model used for voice activity detection, speaker turn detection, overlap detection and resegmentation
    """
    def __init__(self,
                 input_size,
                 blstm_sizes,
                 num_layers):
        """

        :param input_size:
        :param blstm_sizes:
        """
        super(BLSTM, self).__init__()
        self.input_size = input_size
        self.blstm_sizes = blstm_sizes
        self.output_size = blstm_sizes * 2

        self.blstm_layers = torch.nn.LSTM(input_size,
                                          blstm_sizes,
                                          bidirectional=True,
                                          batch_first=True,
                                          num_layers=num_layers)

    def forward(self, inputs):
        """

        :param inputs:
        :return:
        """
        output, h = self.blstm_layers(inputs)
        return output

    def output_size(self):
        """

        :return:
        """
        return self.output_size


def _init_weights(m):
    """

    :return:
    """
    if type(m) == torch.nn.Linear:
        torch.nn.init.xavier_uniform_(m.weight)
        m.bias.data.fill_(0.01)


class SeqToSeq(torch.nn.Module):
    """
    Sequence-To-Sequence model for speech segmentation tasks (VAD, speaker turn detection, overlapped speech detection)
    This model is composed of three main blocks :
        - features extraction : extract audio features from raw audio waveform
        - recurrent layers : reccurent neural network implemented using LSTM (bi-directionnal)
        - MLP : linear layers to project BLSTM outputs to classification domain
    """
    def __init__(self,model_archi):
        """

        :param model_archi: architecture of the model as dictionnary (can be extracted from YAML file)
        """
        super(SeqToSeq, self).__init__()

        cfg = model_archi

        # Get Feature size
        self.feature_size = cfg["feature_size"]
        self.samplerate = cfg["samplerate"]
        self.channel_number = cfg["channel_number"]

        # pre-processing layers
        pre_processing_layers = []
        for k in cfg["pre_processing"].keys():

            if k.startswith("mfcc"):
                n_fft = cfg["pre_processing"][k]["n_fft"]
                win_length = cfg["pre_processing"][k].get("win_length",480)
                hop_length = cfg["pre_processing"][k]["win_shift"]
                n_mels = cfg["pre_processing"][k]["n_mels"]
                n_mfcc = cfg["pre_processing"][k]["n_mfcc"]
                delta = cfg["pre_processing"][k]["delta"]
                delta2 = cfg["pre_processing"][k]["delta-delta"]
                self.center = cfg["pre_processing"][k].get("center",False)
                pad = cfg["pre_processing"][k].get("pad",win_length//2)
                pre_processing_layers.append((k, torchaudio.transforms.MFCC(sample_rate=self.samplerate,
                                                                            n_mfcc=n_mfcc,
                                                                            melkwargs={'n_fft': n_fft,
                                                                                       'n_mels': n_mels,
                                                                                       'hop_length': hop_length,
                                                                                       'win_length':win_length,
                                                                                       'center':self.center,
                                                                                       'pad':pad})))
                pre_processing_layers.append((k+"_delta",Deltas(e=False,
                                                                d=delta,
                                                                dd=delta2)))
                input_size = self.feature_size

            if k.startswith("mel"):
                n_fft = cfg["pre_processing"][k]["n_fft"]
                hop_length = cfg["pre_processing"][k]["win_shift"]
                n_mels = cfg["pre_processing"][k]["n_mels"]
                n_mfcc = cfg["pre_processing"][k]["n_mfcc"]
                pre_processing_layers.append((k, torchaudio.transforms.MelSpectrogram(
                        sample_rate=self.samplerate,
                        n_fft = n_fft,
                        n_mels = n_mels,
                        hop_length = hop_length)))

                input_size = self.feature_size

        self.pre_processing = torch.nn.Sequential(OrderedDict(pre_processing_layers))

        # sequence to sequence network
        sequence_to_sequence_layers = []
        for k in cfg["sequence_to_sequence"].keys():
            if k.startswith("blstm"):
                sequence_to_sequence_layers.append((k, BLSTM(input_size=input_size,
                                                             blstm_sizes=cfg["sequence_to_sequence"][k]["output_size"],
                                                             num_layers=cfg["sequence_to_sequence"][k]["num_layers"])))
                input_size = cfg["sequence_to_sequence"][k]["output_size"] * 2

        self.sequence_to_sequence = torch.nn.Sequential(OrderedDict(sequence_to_sequence_layers))

        # post processing network
        if cfg["activation"] is not None:
            activation = cfg["activation"]
        else:
            activation='tanh'

        if activation == 'tanh':
            post_processing_activation = torch.nn.Tanh()
        elif activation == 'relu':
            post_processing_activation = torch.nn.ReLU()
        elif activation == 'sigmoid':
            post_processing_activation = torch.nn.Sigmoid()
        else:
            post_processing_activation = torch.nn.Tanh()

        post_processing_layers = []
        for k in cfg["post_processing"].keys():

            if k.startswith("lin"):
                post_processing_layers.append((k, torch.nn.Linear(input_size,
                                                            cfg["post_processing"][k]["output"])))
                input_size = cfg["post_processing"][k]["output"]

            elif k.startswith("activation"):
                post_processing_layers.append((k, post_processing_activation))

            elif k.startswith('batch_norm'):
                post_processing_layers.append((k, torch.nn.BatchNorm1d(input_size)))

            elif k.startswith('dropout'):
                post_processing_layers.append((k, torch.nn.Dropout(p=cfg["post_processing"][k])))

        self.post_processing = torch.nn.Sequential(OrderedDict(post_processing_layers))
        self.post_processing.apply(_init_weights)
        self.output_size = input_size

    def forward(self, inputs):
        """

        :param inputs: raw audio signal
        :return:
        """
        if self.center:
            x = self.pre_processing(inputs[:, :, :-1])
        else:
            x = self.pre_processing(inputs[:, :, :])
        # remove energy
        if len(x.shape) == 4:
            x = torch.squeeze(x[:, :, :])
        x = self.sequence_to_sequence(x.permute(0, 2, 1))
        x = self.post_processing(x)
        return x

    def get_output_size(self):
        """

        :return:
        """
        return self.output_size

    def predict(self,
                wav_filename,
                uem_filename,
                device,
                shift,
                output_rate,
                th_in,
                th_out,
                seg_duration=2.0,
                show=None,
                audio_fr=16000,
                out_fr=100,
                batch_size=64,
                task="vad",
                mode="mean"
                ):
        """
            A MODIFIER POU NE PRENDRE QUE LE NOM DU FICHIER WAV ET CRÉER LE DATA LOADER À L'INTERIEUR

        :param seg_duration:
        :param show:
        :param mode:
        :param wav_filename:
        :param batch_size:
        :param uem_filename:
        :param out_fr:
        :param audio_fr:
        :param task:
        :param th_out:
        :param th_in:
        :param output_rate:
        :param shift:
        :param model:
        :param validation_loader:
        :param device:
        :param only_lab_generation: only generating lab file for each wav file without evaluation
        :return: the raw ouput of the network (probability for each class) the vector of labels
        """
        if show is None:
            show = wav_filename

        data_set = SeqSetSingle(show=show,          # Pourquoi????
                                wav_fn=wav_filename,
                                label_set=None,     # Pourquoi???
                                audio_framerate=audio_fr,
                                output_framerate=out_fr,
                                duration=seg_duration,
                                task="vad",
                                uem=uem_filename,   # initalement UEM
                                shift=shift)

        data_loader = DataLoader(data_set,
                                 batch_size=batch_size,
                                 shuffle=False,
                                 drop_last=False,
                                 pin_memory=True,
                                 num_workers=5
                                 )

        output_data = []
        output_idx = []

        sm = torch.nn.Softmax(dim=2)
        with torch.no_grad():
            # TODO : to speed up
            for batch_idx, (win_idx, data) in enumerate(data_loader):
                output = sm(self.forward(data.to(device))).cpu().numpy()
                for ii in range(output.shape[0]):
                    output_data.append(output[ii])
                    output_idx.append(int(win_idx[ii]))

        final_output, _ = multi_label_combination(output_idx,
                                                  None,
                                                  output_data,
                                                  shift,
                                                  output_rate,
                                                  mode=mode)

        if task == "vad":
            prediction = filter_vad(final_output, th_in, th_out)

        elif task == "speaker_turn":
            prediction = filter_speaker_turn(final_output, th_in, output_rate, min_durationi_spkt)
        elif task == "overlap":
            prediction = filter_overlap(final_output, th_in, th_out)
        else:
            raise NotImplementedError()

        return prediction.astype("int"), final_output




# -*- coding: utf-8 -*-
#
# This file is part of s4d.
#
# s4d is a python package for speaker diarization.
# Home page: http://www-lium.univ-lemans.fr/s4d/
#
# s4d is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# s4d is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with s4d.  If not, see <http://www.gnu.org/licenses/>.

"""
Copyright 2014-2021 Anthony Larcher, Théo Mariotte

"""

import itertools
import numpy


def multi_label_combination(output_idx, output_target, output_data, shift, output_rate, mode="mean"):
    """

    :param output_idx:
    :param output_target:
    :param output_data:
    :param shift:
    :param output_rate:
    :param mode:
    :return:

    Author: Anthony Larcher
    """
    win_shift = int(shift * output_rate)

    # Initialize the size of final_output
    final_output = numpy.zeros((win_shift * (len(output_data) - 1) + output_data[0].shape[0], output_data[0].shape[1]))

    final_target = numpy.zeros(win_shift * (len(output_data) - 1) + output_data[0].shape[0])
    overlaping_label_count = numpy.zeros(final_output.shape)

    win_len = output_data[0].shape[0]
    tmp = numpy.ones(output_data[0].shape)

    # Loop on the overlaping windows
    for idx, tmp_d in zip(output_idx, output_data):
        start_idx = win_shift * idx
        stop_idx = start_idx + win_len

        overlaping_label_count[start_idx: stop_idx, :] += tmp
        final_output[start_idx: stop_idx, :] += tmp_d

    # Divide by the number of overlapping values
    final_output /= overlaping_label_count

    if output_target is not None:
        for idx, tmp_t in zip(output_idx, output_target):
            start_idx = win_shift * idx
            stop_idx = start_idx + win_len
            final_target[start_idx: stop_idx] += tmp_t

        final_target /= overlaping_label_count[:, 0].squeeze()

    return final_output, final_target


def filter_vad(signal, th_in, th_out):
    """

    :param signal:
    :param th_in:
    :param th_out:
    :return:
    """
    speech = False
    prediction = numpy.zeros(signal.shape[0], dtype='bool')
    ii = 0
    while ii < signal.shape[0]:
        if signal[ii, 1] > th_in and not speech:
            speech = True
        elif signal[ii, 1] < th_out and speech:
            speech = False
        prediction[ii] = speech
        ii += 1
    return prediction


def filter_speaker_turn(signal, th, output_rate, min_durationi_spkt):
    """

    :param min_durationi_spkt:
    :param output_rate:
    :param th:
    :param signal:
    :return:
    """
    prediction = numpy.zeros(signal.shape[0], dtype='bool')

    order = max(1, int(round(output_rate * min_durationi_spkt, 0)))
    final_output = signal[:, 1].squeeze()
    idx_max = dsp.argrelmax(final_output, order=order)
    idx_max_th = [ii for ii in idx_max[0] if final_output[ii] > th]
    prediction[idx_max_th] = True

    return prediction


def filter_overlap(signal, th_in, th_out):
    """

    :param signal:
    :param th_in:
    :param th_out:
    :return:
    """
    prediction = numpy.zeros(signal.shape[0], dtype='bool')
    overlap = False
    ii = 0
    while ii < signal.shape[0]:
        if signal[ii, 1] > th_in and not overlap:
            overlap = True
        elif signal[ii, 1] < th_out and overlap:
            overlap = False
        prediction[ii] = overlap
        ii += 1

    return prediction


def create_rttm(output_filename, data, show, min_on=0.5, min_off=1):
    """

    :param output_filename:
    :param data:
    :param show:
    :param min_on:
    :param min_off:
    :return:
    """
    with open(output_filename, 'w') as fh:
        ts = 0.01
        curr_time = 0

        grouped = [(x, len(list(y))) for x, y in itertools.groupby(data)]
        first_clean = []
        for x in grouped:
            if x[0] == 0 and x[1] > min_off / ts:
                first_clean.append(x)
            else:
                first_clean.append((1, x[1]))
        expanded_list = list(itertools.chain.from_iterable([[x] * y for x, y in first_clean]))
        grouped2 = [(x, len(list(y))) for x, y in itertools.groupby(expanded_list)]
        second_clean = []
        for x in grouped2:
            if x[0] == 1 and x[1] > min_on / ts:
                second_clean.append(x)
            else:
                first_clean.append((0, x[1]))
        expanded_list2 = list(itertools.chain.from_iterable([[x] * y for x, y in first_clean]))
        grouped3 = [(x, len(list(y))) for x, y in itertools.groupby(expanded_list)]
        for turn in grouped3:
            if turn[0] == 1:
                start = curr_time
                curr_time += (turn[1] * ts)
                stop = turn[1] * ts
                rttm = f"SPEAKER {show} 1 {start:.2f} {stop:.2f} <NA> <NA> OVER <NA> <NA>"
                fh.write(rttm + '\n')
            else:
                curr_time += (turn[1] * ts)

        fh.close()
